#
# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


"""LDAP related stuff for nupyf.

This is a set to handle ldap operations for nupyf:
- connection
- mapping of rules to ldap NuAccessControlList object definition
- sync of ldap acls with rules from xml
"""
__version__ = '0.2'
__author__ = 'Jean Gillaux'

import ldap
import ldap.modlist
from sets import Set
from IPy import IP
import nuacl
import hmac
import time, os
import string
import nubackend


A_CN='cn'
A_OBJ_CLASS='objectClass'

ATTR_TYPE_MULTIPLE=nubackend.ATTR_TYPE_MULTIPLE
ATTR_TYPE_SIMPLE = nubackend.ATTR_TYPE_SIMPLE | Set([A_CN,A_OBJ_CLASS])
ALL_ATTR=(ATTR_TYPE_MULTIPLE | ATTR_TYPE_SIMPLE) 
COMPARE_ATTR=ALL_ATTR - Set([A_CN])

class LDAPConn(object):
    """Handles connection to ldap server, and store information relative to the connection.

    Example:
    conn=LDAPConn('ldap.company','cn=admin,dc=company,dc=com','secret','ou=acls,dc=company,dc=com')
    basedn is the basedn for nufw acces control lists.
    """
    def __init__(self,host,who,cred,basedn,port=''):
	self.host=host
	self.port=port
	self.basedn=basedn
	if self.port!='':
	    self.conn=ldap.open(self.host)
	else: self.conn=ldap.open(self.host)
	self.conn.simple_bind_s(who,cred)

class LDAPTool(object):
    """Handles search, add and delete operations on a ldap tree.    
    """
    def __init__(self,conn,basedn=''):
	self._c=conn
	self._basedn=conn.basedn
	if basedn!='': self._basedn=basedn

    def search(self):
	l=[]
	r=self._c.conn.search(self._basedn,ldap.SCOPE_SUBTREE,'(objectClass=NuAccessControlList)')
	while 1:
	    result_type, result_data=self._c.conn.result(r,0)
	    #print result_data
	    if result_data == []:
		break
	    else:
		lrule=LDAPRule(ldapresult=result_data)
		l.append(lrule)
	return l

    def add(self,ldaprule):
	"""Add an object in the ldap tree.
	
	the dn is made with "cn" atrribute of ldaprule and basedn
	"""
	h=ldaprule.attrs()
	dn='cn='+ldaprule[A_CN][0]+','+self._basedn
	#print 'add ldap',dn	
	modlist=ldap.modlist.addModlist(h)
	#print 'Add_LDAP modlist',modlist
	self._c.conn.add_s(dn,modlist)
	
    def delete(self,ldaprule):
	"""Delete an object from the ldap tree. 
	
	ldaprule is an LDAPRule object
	"""
	dn='cn='+ldaprule[A_CN][0]+','+self._basedn
	#print 'delete ldap',dn
	self._c.conn.delete_s(dn)
	


class LDAPRule(nubackend.NURule):
    def __init__(self,xmlrule=None,ldapresult=None):
    	super(LDAPRule,self).__init__(xmlrule=xmlrule)
	self.h[A_OBJ_CLASS]=Set(['top','NuAccessControlList'])
	self.h[A_CN] = Set([''])
	self.ATTR_TYPE_SIMPLE = ATTR_TYPE_SIMPLE
	self.ATTR_TYPE_MULTIPLE = ATTR_TYPE_MULTIPLE
	self.COMPARE_ATTR = COMPARE_ATTR
	#print self.__cpt
	if xmlrule is not None:
	    self[A_CN] = string.join([xmlrule.elt_src().get('parent_name'),xmlrule.elt_dest().get('parent_name'),xmlrule.proto().get('parent_name'),
				    str(xmlrule.cpt)],'_').encode("utf-8")

	if ldapresult is not None:
	    ldaph = ldapresult[0][1]
	    for k,v in ldaph.iteritems():
		#print '__init__ from ldap k v',k,v
		if k in ATTR_TYPE_SIMPLE:
		    self.h[k] = Set(v)
		if k in ATTR_TYPE_MULTIPLE:
		    self.h[k] = Set(v)
		#print '__init__ from ldap2 k v',k,self.h[k]
        
    def attrs(self):
	res={}
	for k,v in self.h.iteritems():
	    res[k]=[i for i in v]
	return res

        


class LDAPfw(nubackend.NUBackend):
    """This class inserts/removes rules into ldap tree.

    It's ldap nufw backend.
    """
    def __init__(self,conn,fw=None):
	super(LDAPfw,self).__init__(fw=fw,backend_class=LDAPRule)
	self.__ldaprules=[]
	self__conn=conn
	if conn is not None:
	    self.__tool=LDAPTool(conn)
	self.__rcpt = 0
	#
	self.backend_class=LDAPRule

   	
    def add_rule_from_ldap(self,r):
	self.__ldaprules.append(r)
	
    def load_ldap(self):
	self.__ldaprules=self.__tool.search()

    def process(self):
	"""This method takes rules from xml and ldap, then merge them into ldap tree.
	
	- if a rule from ldap is not in the xml rules, it is deleted from the ldap tree.
	- if a rule from xml is not in the ldap rules, it is inserted in the ldap tree.
	"""
	pid = str(os.getpid())
	ppid = str(os.getppid())
	for rxml in self.rules:
	    if rxml not in self.__ldaprules:
	    	old = rxml[A_CN][0]
		new = old + '_%s_%s'%(ppid,pid)
		rxml[A_CN] = new
		self.__tool.add(rxml)
	for rldap in self.__ldaprules:
	    if rldap not in self.rules:
	        self.__tool.delete(rldap)
    #Getters/Setters

    def set_ldap_rules(self,list_of_ldap_rules):
	self.__ldaprules=list_of_ldap_rules
	
    def set_conn(self,conn):
	self.__conn = conn
	self.__tool = LDAPTool(self.__conn)
    
    def ldap_rules(self):
	return self.__ldaprules
