#! /usr/bin/env python

#
# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#


"""Generates iptables commands and ldap nufw acls from xml file.

Nupyf takes an nufw acls xml file and an nufw firewall xml file and can
    * output iptables commands for both nufw and rescue (standard) modes
    * inserts nufw acls into an ldap tree

usage: nupyf [options] acls.xml firewall.xml

"""


__version__ = '1.0-rc3'
__author__ = 'Jean Gillaux'
__copyright__ = 'Copyright 2005, INL'

import xml.dom.minidom
import sys
from nuacl import *
import fw
import ipt
from os import linesep, times
from optparse import OptionParser
from time import strftime
from getpass import getpass
import nuldap
import nunat
from ldap import LDAPError
import cPickle

def check_ldap(options):
    c = 0
    error = 0
    if options.ldap_server !='': c+=1
    if options.ldap_basedn != '': c+=1
    if c!=0 :
	if c!=2: 
	    print >> sys.stderr, 'Error: --server and --basedn options must be provided together'
	    sys.exit(1)
	if options.ldap_basedn =='':
	    print >> sys.stderr, 'Error: basedn must not be empty.'
	    sys.exit(1)

def parse_config_file(config_file,options):
    """Parse config_file and put directives into options.
    """
    try:
	config = eval(open(config_file).read())
    except IOError, ioe:
	print ioe
	sys.exit(1)
    for k,v in config.iteritems():
	if k == 'ldap_password': options.ldap_pwd = v
	if k == 'ldap_user': options.ldap_user = v
	if k == 'ldap_server': options.ldap_server = v
	if k == 'ldap_basedn': options.ldap_basedn = v

def try_write_file(file,*args):
    do_stdout=0
    if file=='-':
        do_stdout=1
    if do_stdout!=1:
        fd=open(file,'w')
    else: fd=sys.stdout
    for arg in args:
	if type(arg) == list:
	    fd.writelines(arg)
	if type(arg) in [str, unicode]:
	    fd.write(arg)
    if do_stdout != 1:
	fd.close()


def main(argv):
    usage = "usage: %prog [options] firewall_file.xml acls_file.xml"
    parser = OptionParser(usage,version='%prog '+ __version__)
    # command line options
    parser.add_option('-d','--dispatch',dest='dispatch',help='file to write dispatch commands in',
    			metavar='FILE',default='')
    parser.add_option('-f','--forward',dest='forward',help='file to write forward iptables rules in', metavar='FILE')
    parser.add_option('-i','--input',dest='input',help='file to ouput input rules in',metavar='FILE',default='')
    parser.add_option('-o','--output',dest='output',help='file to write output rules in',metavar='FILE',default='')
    parser.add_option('-v','--vpn',dest='vpn_rules',help='file to output vpn rules for packet marks, special value - means stdout',
			metavar='FILE',default='')
    parser.add_option('-n','--nat',dest='nat_rules',help='file to output nat rules in, special value - means stdout',
			metavar='FILE')
    parser.add_option('-r','--rescue',dest='rescue',action='store_true',help='activate rescue mode for firewall rules generation')
    parser.add_option('','--ulog',dest='ulog',action="store_true",help='use ULOG system in log commands, default is to use LOG.')
    parser.add_option('','--iptables',dest='ipt',help='path to iptables, can be a variable.',metavar='VALUE',default='')
    parser.add_option('-s','--server',dest='ldap_server',help='ldap server address.',metavar='ADDR',default='')
    parser.add_option('-u','--user',dest='ldap_user',help='ldap user used to connect.',metavar='USER',default='')
    parser.add_option('-p','--pwd',dest='ldap_pwd',help='password used to connect on ldap server (WARNING: visible on comand line).',
			metavar='PWD',default='')
    parser.add_option('-b','--basedn',dest='ldap_basedn',help='basedn used to insert acls in (ex: ou=acls,dc=domain,dc=com).',
			metavar='DN',default='')
    parser.add_option('-a','--askpwd',dest='ask_ldappwd',help='ask user for ldap passwd.',action='store_true',default=False)
    parser.add_option('-c','--config',dest='config_file',help='configuration file',default='')
    parser.add_option('','--dumpldap',dest='dump_ldap',help='file to save ldap informations',default='')
    parser.add_option('','--loadldap',dest='load_ldap',help='file to load ldap informations from',default='')
    parser.add_option('','--auth_ext',dest='auth_ext',action='store_true',help='Generate netfilter rules to autenticate internet')
    
    parser.set_defaults(rescue=False,forward='',nat_rules='',auth_ext=False)
    (options,args)=parser.parse_args(argv)
    if options.config_file != '':
	parse_config_file(options.config_file,options)
    #print 'ldap: ',options.ldap_server,options.ldap_user,options.ldap_pwd,options.ldap_basedn
    #sys.exit(0)
    if options.ask_ldappwd:
	options.ldap_pwd = getpass('LDAP Password: ')    		    		
    do_ldap = False
    check_ldap(options)
    if options.ldap_server !='' and options.ldap_basedn != '':
	do_ldap = True
    if options.load_ldap != '':
	dapfw = cPickle.load(open(options.load_ldap))
	ldapconn=nuldap.LDAPConn(options.ldap_server,options.ldap_user,options.ldap_pwd,options.ldap_basedn)
	dapfw.set_conn(ldapconn)
	dapfw.load_ldap()
	dapfw.process()
	return 0
    
    if options.ipt!='':# set iptables path 
	#print 'setting ipt'
	ipt.IPTABLES=options.ipt
	#print ipt.IPTABLES
    do_stdout=0
    log_type='log'
    if options.ulog: log_type='ulog'
        
    doc = xml.dom.minidom.parse(args[2])

    # parse groups
    groups_list = parse_groups(doc)
    if not groups_list.has_key(1): groups_list[1] = 1 #group 1 is enabled by default
    #print "groups: ",groups_list
    #sys.exit(0) #FIXME: debug exit
    lsnat,ldnat,lpnat = nunat.nats_from_xml(doc,groups_list)
    #print lsnat;
    lr = doc.getElementsByTagName("ressource")
    ls = doc.getElementsByTagName("subject")
    lp = doc.getElementsByTagName("protocol")
    lapp = doc.getElementsByTagName("application")
    listr=[]
    listrobj=EltGrpList()
    lists=[]
    listsobj=EltGrpList()
    listp=[]
    listpobj=EltGrpList()
    listapp=[]
    listappobj=EltGrpList()

    for r in lr:
        #print r
	elts=r.getElementsByTagName("elt")
        #print elts
	try:
	    mygrp=xml_new_eltgrp(r.getAttribute('ID'),r.getAttribute('name'),'or',elts)
	except ValueError, e:
	    print e
	    return 1
        #print 'new ressource ',mygrp
	listr.append(mygrp)
        listrobj.add(mygrp)

    for s in ls:
	elts2=s.getElementsByTagName("elt")
	try:
	    mygrp2=xml_new_eltgrp(s.getAttribute('ID'),s.getAttribute('name'),s.getAttribute('op'),elts2)
	except ValueError, e:
	    print e
	    return 1
	lists.append(mygrp2)
        listsobj.add(mygrp2)

    for pr in lp:
	elts3=pr.getElementsByTagName("elt")
	try:
	    mygrp3=xml_new_eltgrp(pr.getAttribute('ID'),pr.getAttribute('name'),'or',elts3)
	except ValueError, e:
	    print e
	    return 1
	listp.append(mygrp3)
        listpobj.add(mygrp3)
	
    for pr in lapp:
	elts3=pr.getElementsByTagName("elt")
	try:
	    mygrp4=xml_new_eltgrp(pr.getAttribute('ID'),pr.getAttribute('name'),pr.getAttribute('op'),elts3)
	    #print mygrp4, mygrp4.eltset_t
	except ValueError, e:
	    print e
	    return 1
	listapp.append(mygrp4)
        listappobj.add(mygrp4)

    #listsobj.make_graph()
    #listpobj.make_graph()
    #listrobj.make_graph()
    #do the operator and link operations
    #Shazam
    listsobj.normall()
    listrobj.normall()
    listpobj.normall()
    listappobj.normall()
    
    lacls=doc.getElementsByTagName("acl")
    l_acls=[]
    #parse network description file, and take first firewall
    fwlist = fw.parse(args[1])
    try:
        (fwid, myfw) = fwlist.popitem()
    except KeyError:
        print "unable to find a firewall"
        sys.exit(1)
    # do we have to authenticate packets from internet?
    myfw.auth_ext = options.auth_ext
        
    for a in lacls:
	src=a.getAttribute("from")
        dst=a.getAttribute("to")
	proto=a.getAttribute("proto")
	idappli=a.getAttribute("with")
	aclname=a.getAttribute("name")
        acl_group = 0
        try:
            acl_group = int(a.getAttribute("group"))
        except ValueError:
            acl_group = None
	#print src+"->"+dst+" with "+proto
	toto = listappobj.find(idappli)
	#print toto, toto.eltset_t
	if idappli != '': appli = listappobj.find(idappli)
	else: appli = None
        #get acl only if its group is enabled
        if not acl_group or (groups_list.has_key(acl_group) and groups_list[acl_group]==1):
            acl=Acl(listsobj.find(src),listrobj.find(dst),myfw,id=a.getAttribute('ID'),proto=listpobj.find(proto),name=aclname,
		decision=a.getAttribute('decision'),appli=appli)

	    l_acls.append(acl)
    
    #net_ent=myfw.entreprise_nets()
    myfw.chains(all=1,inout=1)
    
    for x in l_acls: 
	for r in x.rules():
	    myfw.add_rule(r)
    myfw.set_nat_rules(lsnat, ldnat, lpnat)
    if options.input == '': myfw.manage_input = False
    if options.output == '': myfw.manage_output = False

    #
    # Create iptables commands
    #
    fwp=ipt.FWipt(myfw,logtype=log_type)
    rules_create = fwp.create_ipt_chains()
    rules_connect = fwp.connect_chains(rescue=options.rescue)

    input_rules = output_rules = srules = []
    input_rules, output_rules, srules = fwp.gen_rules(rescue=options.rescue)
    intro="""
#Generated by nupyf on %s

""" %(strftime('%d %m %Y at %H:%M'))
    if options.dispatch != '':
	try_write_file(options.dispatch,intro,'#DISPATCH and DEFAULT Rules%s'%(linesep),rules_create,rules_connect)
    if options.forward != '':
	try_write_file(options.forward,intro,'#Rules for FORWARD%s'%linesep,srules)
    if options.input != '':
	try_write_file(options.input,intro,'#Rules for INPUT%s'%linesep,input_rules)
    if options.output != '':
	try_write_file(options.output,intro,'#Rules for OUTPUT%s'%linesep,output_rules)
    if options.vpn_rules != '':
	try_write_file(options.vpn_rules,intro,'#VPN Rules%s'%linesep,fwp.gen_vpn_rules())
    if options.nat_rules != '':
	try_write_file(options.nat_rules,intro,'#NAT Rules%s'%linesep,fwp.gen_nat_rules())


    #LDAP stuff
    if do_ldap:
	conn = None
	ldapfw=nuldap.LDAPfw(conn,fw=myfw)
	if options.dump_ldap != '':
	    cPickle.dump(ldapfw,open(options.dump_ldap,'w+'))
	    return 0
	conn=nuldap.LDAPConn(options.ldap_server,options.ldap_user,options.ldap_pwd,options.ldap_basedn)
	ldapfw.set_conn(conn)
	ldapfw.load_ldap()
	ldapfw.process()
    return 0
    

if __name__ == "__main__":
    #use psyco if available   
    try:
        import psyco
        psyco.full()
    except ImportError:
        pass
    try:
    	res=main(sys.argv)
    except LinkError, linke:
	print >> sys.stderr, str(linke)
	sys.exit(1)
    except LoopError, loope:
	print >> sys.stderr, str(loope)
	sys.exit(1)
    except fw.UnknownVpn, evpn:
	print >> sys.stderr, 'ERROR: unknown vpn with mark %s detected!'%(evpn.mark)
	sys.exit(1)
    except LDAPError, ldaperr:
	print >> sys.stderr, 'ERROR ldap:', ldaperr
	sys.exit(1)
    #except ValueError, ve:
	#print 'main intercept ValueError'
	#print ve
        #print sys.last_traceback
	#sys.exit(1)
    sys.exit(res)
